import os
import sys
import configparser
from pathlib import Path


if (_config_home := os.path.expandvars('$XDG_CONFIG_HOME')) != '$XDG_CONFIG_HOME':
	CONFIG_HOME = Path(_config_home)
else:
	CONFIG_HOME = Path.home() / '.config'


KONFSAVE_DATA_PATH = CONFIG_HOME / 'konfsave'
KONFSAVE_PROFILE_HOME = KONFSAVE_DATA_PATH / 'profiles'
KONFSAVE_PROFILE_INFO_FILENAME = '.konfsave_profile'
KONFSAVE_CURRENT_PROFILE_PATH = Path.home() / KONFSAVE_PROFILE_INFO_FILENAME
KONFSAVE_PRINT_COPYINGED_FILES = False
KONFSAVE_DEFAULT_CONFIG_PATH = Path(__file__).parent / 'default_config.ini'

PATHS_TO_SAVE = set()

# Create the config file if missing
if not (KONFSAVE_DATA_PATH / 'config.ini').exists():
	print('Config file missing, copying from default')
	KONFSAVE_DATA_PATH.mkdir(parents=True, exist_ok=True)
	with open(KONFSAVE_DATA_PATH / 'config.ini', 'w') as f, open(KONFSAVE_DEFAULT_CONFIG_PATH) as d:
		f.write(d.read())

# Load the config file
with open(KONFSAVE_DATA_PATH / 'config.ini') as f:
	config = configparser.ConfigParser(allow_no_value=True)
	config.read_file(f)
	try:
		xdg_paths = list(config['XDG_CONFIG_HOME Paths To Save'].keys())
		home_paths = list(config['Home Directory Paths To Save'].keys())
	except KeyError:
		sys.stderr.write(f'Warning: malformed config file at {KONFSAVE_DATA_PATH / "config.ini"}\n')
		raise
	PATHS_TO_SAVE |= set(map(lambda p: CONFIG_HOME / p, xdg_paths))
	PATHS_TO_SAVE |= set(map(lambda p: Path.home() / p, home_paths))



import sys
import argparse
import subprocess
from pathlib import Path
from typing import List

import constants
import profiles

_N_T = '\n  '  # Backslashes are not allowed in f-string expressions, so use a variable
HELP_TEXT = '''Konfsave is a KDE config manager.

usage: konfsave <action> [args ...] [flags ...]

Actions:
help, --help, -h    print this message and exit
i, info             get info about the current configuration, or a profile if specified
s, save             save the current configuration
l, load             load a saved profile
c, change           modify a profile's attributes
list-files          list files that save would copy

To see detailed usage instructions, run `konfsave <action> --help`.
All flags starting with '--' can be abbreviated.
'''
def parse_arguments(argv):
	# Actions need to parsed separately because each action has its own
	# argument format, so different ArgumentParser objects have to be used.
	if argv[0] == 'python':
		argv = argv[1:]
	action = argv[1] if len(argv) > 1 else 'help'
	try:
		next(v for k, v in {
            ('-h', '--help', 'help'): lambda *_: print(HELP_TEXT),
			('i', 'info'): action_info,
			('list-files'): action_list_files,
			('s', 'save'): action_save,
			('l', 'load'): action_load,
			('c', 'change'): action_change,
			('d', 'delete'): action_delete
		}.items() if action in k)(argv[2:])
	except StopIteration:
		sys.stderr.write(f'Unrecognized action: {action}\nTry \'konfsave help\' for more info.\n')
	

def action_info(argv):
	parser = argparse.ArgumentParser(prog='konfsave info')
	parser.add_argument(
		'profile', nargs='?', metavar='profile_name',
		help='If provided, prints detailed information about the profile.'
	)
	args = parser.parse_args(argv)
	if profile := args.profile:
		profiles.validate_profile_name(profile)
		info = profiles.profile_info(profile)
		if info is None:
			print(f'The profile {profile} doesn\'t exist.')
		else:
			print(f'Name: {info["name"]}')
			print(f'Stored at: {constants.KONFSAVE_PROFILE_HOME / profile}')
			if include := info['include']:
				print(f'Files to additionally include: \n  {_N_T.join(map(str, include))}')
			if exclude := info['exclude']:
				print(f'Files to exclude by default: \n  {_N_T.join(map(str, exclude))}')
	else:
		if current_profile := profiles.current_profile():
			print(f'Current profile: {current_profile}')
			print(f'Stored at: {constants.KONFSAVE_PROFILE_HOME / current_profile}')
		else:
			print(f'No profile is currently active.')
		if saved_profiles := sorted(list(map(lambda q: q.name, filter(
			lambda p: (p / constants.KONFSAVE_PROFILE_INFO_FILENAME).exists(),
			constants.KONFSAVE_PROFILE_HOME.glob('*')
		))), key=str.lower):
			print(f'Saved profiles:\n  {_N_T.join(saved_profiles)}')
		else:
			print('No profiles are saved.')
	
	
def action_list_files(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave list',
		description='Print a list of files that would be saved by \'konfsave save\'.'
	)
	parser.add_argument(
		'profile', nargs='?', default=profiles.current_profile(), metavar='profile_name',
		help='If provided, lists files that would be saved to a specific profile (according to its configuration).'
	)
	args = parser.parse_args(argv)
	if args.profile:
		profiles.validate_profile_name(args.profile)
	try:
		print('\n'.join(sorted(map(str, profiles.paths_to_save(args.profile)), key=str.lower)))
	except ValueError as e:
		print(str(e))


def action_save(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave save',
		# The default usage string puts "[name]" at the end, for some reason.
		usage='konfsave save [-h] [name] [--destination DEST] [--follow-symlinks] [--include [FILE ...]] [--exclude [FILE ...]]'
	)
	parser.add_argument(
		'profile', metavar='name', nargs='?', default=profiles.current_profile(),
		help='Save as the specified profile instead of using the currently loaded name. This is required if no profile is active.'
	)
	parser.add_argument(
		'--destination', metavar='DEST', type=Path,
		help='Instead of saving to Konfsave\'s profile storage, save to a specified destination.'
	)
	parser.add_argument(
		'--follow-symlinks', '-s', action='store_true', dest='follow_symlinks',
		help='By default, symlinks are copied as symlinks. If this flag is used, the contents of symlinked files will be copied.'
	)
	parser.add_argument(
		'--include', action='extend', nargs='*', metavar='FILE', default=[],
		help='Files to add to the profile. Paths must be either absolute or relative to the home directory. '
		'In either case, the actual file must be inside of the home directory.\n'
		'Files specified here will be included even if the profile excludes them by default.'
	)
	parser.add_argument(
		'--exclude', action='extend', nargs='*', metavar='FILE', default=[],
		help='Files to exclude from the profile; that is, they will not be copied,\n'
		'but if they already exist in the profile, they will not be deleted. Path format is the same as for --include. '
		'Files specified here will be excluded even if the profile includes them by default.'
	)
	args = parser.parse_args(argv)
	if args.profile:
		profiles.validate_profile_name(args.profile)
	include = set()
	exclude = set()
	for path in map(Path, args.exclude):
		if not path.is_absolute():
			path = Path.home() / path
		exclude.add(path)
	for path in map(Path, args.include):
		if not path.is_absolute():
			path = Path.home() / path
		include.add(path)
	profiles.save(
		name=args.profile,
		destination=args.destination,
		follow_symlinks=args.follow_symlinks,
		include=include,
		exclude=exclude
	)
	print('Success')
	
	
def action_load(argv):
	parser = argparse.ArgumentParser(prog='konfsave load')
	parser.add_argument(
		'profile',
		help='The name of the profile to load.'
	)
	parser.add_argument(
		'--overwrite',
		help='By default, loading will fail if no profile is active (i.e. the current configuration is not saved). '
		'Otherwise, the user will be asked for confirmation. Using --overwrite will bypass both of these checks.'
	)
	parser.add_argument(
		'--no-restart', action='store_false', dest='restart',
		help='After loading a profile, the Plasma shell is restarted, unless this flag was specified.'
	)
	parser.add_argument(
		'--include', action='extend', nargs='*', metavar='FILE', default=[],
		help='Files to load from the profile in addition to those loaded by default. '
		'Paths must point to their final destination and be either absolute or relative to the home directory. '
		'Files that are loaded by default can be checked using `konfsave list <profile>`.'
	)
	parser.add_argument(
		'--exclude', action='extend', nargs='*', metavar='FILE', default=[],
		help='Files to not load from the profile. This overrides values specified by the profile\'s configuration. '
		'Path format is the same as for --include.'
	)
	args = parser.parse_args(argv)
	profiles.validate_profile_name(args.profile)
	include = set()
	exclude = set()
	for path in map(Path, args.exclude):
		if not path.is_absolute():
			path = Path.home() / path
		exclude.add(path)
	for path in map(Path, args.include):
		if not path.is_absolute():
			path = Path.home() / path
		include.add(path)
	success = not profiles.load(
		args.profile,
		include,
		exclude,
		overwrite_unsaved_configuration=args.overwrite
	)
	if success:
		if args.restart:
			# Can't use --replace because it can't be passed to kstart5.
			subprocess.run(['killall', 'plasmashell'])
			subprocess.run(['kstart5', 'plasmashell'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
		print('Success')
	
	
def action_change(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave change',
		epilog='Paths must be specified relative to the home directory.'
	)
	parser.add_argument(
		'profile', nargs='?',
		help='Profile to change. If not specified, the current profile will be used. '
		'Note that this will change the profile\'s saved attributes as well; '
		'if you\'d like to store new attributes as a separate profile, save the configuration under a new name first.'
	)
	parser.add_argument(
		'--name', help='New name for the profile.'
	)
	parser.add_argument(
		'--include', action='extend', nargs='*', metavar='FILE', default=[],
		help='New set of files to save to this profile in addition to those saved by default.'
	)
	parser.add_argument(
		'--exclude', action='extend', nargs='*', metavar='FILE', default=[],
		help='New set of files to NOT save to this profile even if they\'re saved by default.'
	)
	args = parser.parse_args(argv)
	results = {k: v for k, v in vars(args).items() if k != 'profile' and v}
	if results:
		profiles.change(results, args.profile)
		print('Success')
	else:
		print('Nothing to change')


def action_delete(argv):
	parser = argparse.ArgumentParser(
		prog='konfsave delete',
		description='Delete a profile. Current system configuration will be unchanged; '
		'however, if the deleted profile is active, there will no longer be an active profile.'
	)
	parser.add_argument(
		'profile', help='The profile to delete.'
	)
	parser.add_argument('--noconfirm', action='store_false', dest='confirm')
	args = parser.parse_args(argv)
	success = not profiles.delete(args.profile, confirm=args.confirm)
	if success:
		print('Success')



import sys
import os
import shutil
import json
from pathlib import Path
from typing import Optional, Set

import constants
# TODO: write proper docstrings so that this can be used as a library

def validate_profile_name(name, exit_if_invalid=True) -> bool:
	valid = name.isidentifier()
	if (not valid) and exit_if_invalid:
		sys.stderr.write(f'The profile name "{name}" is invalid.\n')
		sys.exit(1)
	return valid


def copy_path(source, destination, overwrite=True, follow_symlinks=False):
	if constants.KONFSAVE_PRINT_COPYINGED_FILES:
		print(f'Copying {path}')
	destination.parent.mkdir(parents=True, exist_ok=True)
	if source.is_dir():
		shutil.copytree(
			src=source,
			dst=destination,
			symlinks=follow_symlinks,
			copy_function=shutil.copy,
			dirs_exist_ok=overwrite
		)
	else:
		if overwrite or not destination.exists():
			shutil.copy(source, destination, follow_symlinks=follow_symlinks)
			
			
def current_profile():
	try:
		return profile_info()['name']
	except TypeError:
		return None


def paths_to_save(name=None, include=None, exclude=None, default_include=None, nonexisting_ok=False) -> Set[Path]:
	"""
	The name is not validated in this function.
	
	``include``, ``exclude``, and ``default_include`` must be given as absolute paths
	Paths are returned as absolute and resolved
	If ``name`` is not specified, the current profile will be used if one is active;
	otherwise, the default list from config.ini is used
	"""
	default_include = set(map(lambda p: Path(p).resolve(), default_include or constants.PATHS_TO_SAVE))
	include = set(map(lambda p: Path(p).resolve(), include or ()))
	exclude = set(map(lambda p: Path(p).resolve(), exclude or ()))
	if not name:
		info = profile_info()
	else:
		info = profile_info(name)
		if info is None and not nonexisting_ok:
			raise ValueError(f'The profile "{name}" doesn\'t exist.')
	info = info or {'exclude': set(), 'include': set()}
	exclude = (exclude | set(map(lambda p: (Path.home() / p).resolve(), info['exclude']))) - include
	include = (include | set(map(lambda p: (Path.home() / p).resolve(), info['include']))) - exclude
	return (default_include | include) - exclude


def save(name=None, include=None, exclude=None, follow_symlinks=False, destination=None):
	"""
	The name is not validated in this function.
	
	If ``name`` is unspecified, the current profile's name is used.
	``include`` and ``exclude`` must be given in the same format as to ``paths_to_save()``.
	"""
	info = profile_info(name)
	if name is None:
		if info is None:
			raise RuntimeError('Attempted to save the current profile, but no profile is active.')
		else:
			name = info['name']
	profile_dir = (constants.KONFSAVE_PROFILE_HOME / name) if destination is None else destination
	profile_dir.mkdir(parents=True, exist_ok=True)
	for path in map(Path, paths_to_save(name, include, exclude, nonexisting_ok=True)):
		if not path.exists():
			sys.stderr.write(f'Warning: this path doesn\'t exist. Skipping\n')
		elif not path.is_relative_to(Path.home()):
			sys.stderr.write(f'Warning: this path is not within the user\'s home directory. Skipping\n')
		else:
			copy_path(path, profile_dir / path.relative_to(Path.home()), follow_symlinks=follow_symlinks)
	# TODO: implement git repos in profiles
	new_info = {
		'name': name,
		'include': list(info['include']) if info else [],
		'exclude': list(info['exclude']) if info else []
	}
	with open(profile_dir / constants.KONFSAVE_PROFILE_INFO_FILENAME, 'w') as f:
		json.dump(new_info, f)


def load(name, include=None, exclude=None, overwrite_unsaved_configuration=False) -> bool:
	"""
	The name is not validated in this function.
	True is returned if the user canceled the action.
	
	The KDE configuration will be overwritten if:
		* ``overwrite_unsaved_configuration`` is True
			OR all of the following is true:
		* A profile is active
		* The source profile's info JSON is valid
		* The user manually confirmed that they want to overwrite their configuration
	"""
	profile_root = constants.KONFSAVE_PROFILE_HOME / name
	if overwrite_unsaved_configuration is not True:  # Be really sure that overwriting is intentional
		# If the checks below fail, exit the function.
		try:
			current_info = profile_info()
			assert (profile_root / constants.KONFSAVE_PROFILE_INFO_FILENAME).is_file()
			if current_info is None:
				print('Warning: there is no active profile, and the current configuration is NOT SAVED.')
			elif current_info['name'] != name:
				print('Warning: you\'re loading a new profile. Konfsave cannot check if the currently active profile has been updated with current files.')
			else:
				print('Warning: you\'re overwriting the current profile with existing configuration. Konfsave cannot check if the saved files are up-to-date.')
			if input('Are you sure you want to overwrite the current system configuration? [y/N]: ').lower() != 'y':
				print('Loading aborted.')
				return True
		except Exception as e:
			sys.stderr.write('Refusing to overwrite unsaved configuration\n')
			raise
	constants.KONFSAVE_CURRENT_PROFILE_PATH.unlink(missing_ok=True)
	for path in map(lambda p: Path(p).relative_to(Path.home()), paths_to_save(name, include, exclude)):
		source = profile_root / path
		if source.exists():
			copy_path(source, Path.home() / path)
		else:
			sys.stderr.write(f'Warning: the file {source} doesn\'t exist. Skipping\n')
	shutil.copyfile(profile_root / constants.KONFSAVE_PROFILE_INFO_FILENAME, constants.KONFSAVE_CURRENT_PROFILE_PATH)


def change(results, profile=None):
	"""
	``results`` is a dictionary containing modified attributes and their values.
	Attributes that exist in the profile but are not present in ``results`` will be kept as is.
	
	If the name is changed, ``rename()`` will be called. Additionally, if ``profile`` is
	the same as the current profile, the current profile's attributes will be changed as well.
	"""
	current = current_profile()
	if profile is None:
		if current is None:
			raise RuntimeError('Attempted to change the current profile, but no profile is active.')
		profile = current
	new_info = profile_info(profile, convert_sets=False).copy()
	new_info.update(results)
	if profile == current:
		# Also modify the profile info stored in the home directory
		with open(constants.KONFSAVE_CURRENT_PROFILE_PATH, 'w') as f:
			f.write(json.dumps(new_info))
	if 'name' in results:
		rename(profile, results['name'], change_info=False)  # Avoid writing to the file twice
	with open(constants.KONFSAVE_PROFILE_HOME / new_info['name'] / constants.KONFSAVE_PROFILE_INFO_FILENAME, 'w') as f:
		f.write(json.dumps(new_info))  # Write only after JSON serialization is successful
		
		
def rename(source, result, change_info=True):
	"""
	Rename a saved profile by changing its info and directory. Only the resulting name is validated.
	The current profile is not modified even if its name matches the source name.
	"""
	validate_profile_name(result)
	if (constants.KONFSAVE_PROFILE_HOME / result).exists():
		raise FileExistsError(f'A profile named "{result}" is already saved.')
	if not (constants.KONFSAVE_PROFILE_HOME / source).exists():
		raise RuntimeError(f'The profile "{source}" doesn\'t exist.')
	if change_info:
		info = profile_info(source)
		info.update({'name': result})
		with open(constants.KONFSAVE_PROFILE_HOME / source / constants.KONFSAVE_PROFILE_INFO_FILENAME, 'w') as f:
			f.write(json.dumps(new_info))  # Write only after JSON serialization is successful
	(constants.KONFSAVE_PROFILE_HOME / source).rename(constants.KONFSAVE_PROFILE_HOME / result)


def delete(profile, clear_active=True, confirm=True) -> bool:
	"""
	If ``clear_active`` is True and the deleted profile has the same name as the active profile,
	the active profile info will be deleted as well. Current configuration will be unaffected.
	
	True is returned if the user canceled the action.
	"""
	if confirm:
		print(f'Warning: you\'re about to delete the profile "{profile}".')
		if input('Are you sure you want to permanently delete it? [y/N]: ') != 'y':
			print('Deleting aborted.')
			return True
	if clear_active and profile == current_profile():
		constants.KONFSAVE_CURRENT_PROFILE_PATH.unlink(missing_ok=True)
	shutil.rmtree(constants.KONFSAVE_PROFILE_HOME / profile)


def profile_info(profile_name=None, convert_sets=True) -> Optional[dict]:
	"""
	If the profile name is invalid, this function will print a warning and continue normally.
	
	When no argument is supplied, this will read ``constants.KONFSAVE_CURRENT_PROFILE_PATH``.
	The return value is ``None`` if the JSON file is missing or malformed.
	"""
	if profile_name and not validate_profile_name(profile_name, exit_if_invalid=False):
		sys.stderr.write(f'Warning: "f{profile_info}" is an invalid profile name\n')
	try:
		return parse_profile_info(
			(constants.KONFSAVE_PROFILE_HOME / profile_name / constants.KONFSAVE_PROFILE_INFO_FILENAME) \
				if profile_name else constants.KONFSAVE_CURRENT_PROFILE_PATH, convert_sets=convert_sets
		)
	except FileNotFoundError:
		return None


def parse_profile_info(profile_info_file_path, convert_sets=True) -> Optional[dict]:
	try:
		with open(profile_info_file_path) as f:
			info = json.load(f)
			assert info['name'].isidentifier()
			info['include'] = map(Path, info['include'] or ())
			info['exclude'] = map(Path, info['exclude'] or ())
			if convert_sets:
				info['include'] = set(info['include'])
				info['exclude'] = set(info['exclude'])
			else:
				info['include'] = list(info['include'])
				info['exclude'] = list(info['exclude'])
			return info
	except (json.JSONDecodeError, KeyError, AssertionError) as e:
		sys.stderr.write(f'Warning: malformed profile info at {profile_info_file_path}\n{str(e)}\n')
		return None
	
